from typing import Type, cast


class SentinelMCS(type):
    def __repr__(self):
        return f"<sentinel class '{self.__name__}'>"


class Sentinel(metaclass=SentinelMCS):
    def __new__(cls):
        inst = super().__new__(cls)
        cls.__new__ = lambda _: inst
        return inst

    def __repr__(self):
        return f"<Sentinel:{type(self).__qualname__}>"


def sentinel(name: str) -> Type[Sentinel]:
    """Create a new sentinel type."""

    return cast(Type[Sentinel], SentinelMCS(name, (Sentinel,), {}))



from typing import Generic, NoReturn, TypeVar, Type

from tramp.singleton import singleton

V = TypeVar("V")


class OptionalException(Exception):
    """Base exception for errors raised by an Optional object."""


class OptionalTypeCannotBeInstantiated(OptionalException):
    """Raised when attempting to instantiate an Optional type that is not a type of Some."""


class OptionalHasNoValueException(OptionalException):
    """Raised when an Optional object has no value."""


class Optional(Generic[V]):
    Some: "Type[Optional[V]]"
    Nothing: "Type[Optional[V]]"

    def __new__(cls, *_):
        if cls is Optional:
            raise OptionalTypeCannotBeInstantiated(
                "You cannot instantiate the base Optional type."
            )

        return super().__new__(cls)

    @property
    def value(self) -> V | NoReturn:
        raise OptionalHasNoValueException(
            "You cannot access Optional directly, you must use either Optional.Some or Optional.Nothing"
        )

    def value_or(self, default: V) -> V:
        return self.value

    def __bool__(self):
        return False


class Some(Optional):
    __match_args__ = ("value",)

    def __init__(self, value: V):
        self._value = value

    @property
    def value(self) -> V:
        return self._value

    def __bool__(self):
        return True


@singleton
class Nothing(Optional):
    @property
    def value(self) -> NoReturn:
        raise OptionalHasNoValueException("No value was set, this is Nothing")

    def value_or(self, default: V) -> V:
        return default


Optional.Some = Some
Optional.Nothing = Nothing



from typing import TypeVar, Type

T = TypeVar('T', bound='Singleton')


def singleton(cls: Type[T]) -> Type[T]:
    """A decorator to make a class a singleton."""
    old_new = cls.__new__
    def new(*args):
        inst = old_new(*args)
        cls.__new__ = lambda *_: inst
        return inst

    cls.__new__ = new
    return cls



from typing import Generic, NoReturn, TypeVar, Type

V = TypeVar("V")


class ResultException(Exception):
    """Base exception for errors raised by a result object."""


class ResultTypeCannotBeInstantiated(ResultException):
    """Raised when attempting to instantiate a result type that is not a value type."""


class ResultWasAnErrorException(ResultException):
    """Raised when a result object wraps an error."""


class ResultWasNeverSetException(ResultException):
    """Raised when a result is never given a value and no errors were raised."""


class ResultBuilder(Generic[V]):
    def __init__(self):
        self._result: Result[V] | None = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self._result = Result.Error(exc_val)

        elif self._result is None:
            raise ResultWasNeverSetException("No value was ever set on the result.")

        return True

    @property
    def result(self) -> "Result[V] | NoReturn":
        return self._result

    def set(self, value: V):
        self._result = Result.Value(value)


class Result(Generic[V]):
    Value: "Type[Result[V]]"
    Error: "Type[Error[V]]"

    def __new__(cls, *_):
        if cls is Result:
            raise ResultTypeCannotBeInstantiated(
                "You cannot instantiate the base result type."
            )

        return super().__new__(cls)

    def __bool__(self):
        return False

    @property
    def value(self) -> V | NoReturn:
        raise RuntimeError

    def error(self) -> Exception | None:
        return

    def value_or(self, default: V) -> V:
        pass

    @classmethod
    def build(cls) -> ResultBuilder[V]:
        return ResultBuilder()


class Value(Result[V]):
    __match_args__ = ("value",)

    def __init__(self, value: V):
        self._value = value

    def __repr__(self):
        return f"{Result.__name__}.{type(self).__name__}({self._value!r})"

    def __bool__(self):
        return True

    @property
    def value(self) -> V:
        return self._value


class Error(Result[V]):
    __match_args__ = ("error",)

    def __init__(self, error: Exception):
        self._error = error

    def __repr__(self):
        return f"{Result.__name__}.{type(self).__name__}({self._error!r})"

    @property
    def value(self) -> NoReturn:
        raise ResultWasAnErrorException("The result was an error.") from self.error

    @property
    def error(self) -> Exception:
        return self._error

    def value_or(self, default: V) -> V:
        return default


Result.Value = Value
Result.Error = Error



from typing import Type, Any, get_args, get_origin, Annotated, Union
from collections.abc import Mapping
from tramp.optionals import Optional


class Annotation:
    """Tramp's Annotation type allows for the evaluation of annotations as needed. It's lazy and is lax about names to
    allow for forward references. It takes an annotation that can be a string, a type, or any object along with the
    namespace to evaluate the annotation in.

    If the annotation is a string, it will be evaluated in the namespace using an EvaluationNamespace to prevent all
    name errors. Anytime the annotation references a name that is not in the namespace, a AnnotationForwardReference will be
    injected in its place.

    Evaluations happen every time that the annotation is accessed. This allows for the annotation to be evaluated in
    realtime, reflecting any changes in the namespace.

    A simple example
        import sys
        from inspect import get_annotations

        from tramp.annotations import Annotation
        from tramp.modules import get_module_namespace
        from tramp.optionals import Optional

        class Example:
            foo: "list[Bar]"

        annotation = Annotation(get_annotations(Example)["foo"], Optional.Some(get_module_namespace(__name__, True))
        print(annotation.type)  # <class 'list'>
        print(annotation.args)  # (<AnnotationForwardReference: 'Bar'>,)

        class Bar:
            ...

        print(annotation.type)  # <class 'list'>
        print(annotation.args)  # (<class 'Bar'>,)
    """
    def __init__(self, annotation: str | Type[Any] | Any, some_namespace: Optional[dict[str, Any]] = Optional.Nothing):
        self._annotation = annotation
        self._some_namespace = some_namespace

    @property
    def annotation(self) -> Type[Any] | Any:
        return self._get_annotation()

    @property
    def args(self) -> tuple[Type[Any] | Any, ...]:
        args = get_args(self.annotation)
        if self.is_annotated():
            return args[1:]

        return args

    @property
    def origin(self) -> Type[Any] | Any:
        return get_origin(self.annotation)

    @property
    def type(self) -> Type[Any]:
        if self.is_annotated():
            return Annotation(get_args(self.annotation)[0], self._some_namespace).type

        if self.is_generic():
            return self.origin

        if self.is_type():
            return self.annotation

        if self.args:
            return self.args[0]

        return self.annotation

    def is_annotated(self) -> bool:
        return self.origin is Annotated

    def is_generic(self) -> bool:
        return isinstance(self.origin, type)

    def is_optional(self) -> bool:
        return self.is_union() and type(None) in self.args

    def is_type(self) -> bool:
        return isinstance(self.annotation, type)

    def is_union(self) -> bool:
        return self.origin is Union

    def _get_annotation(self) -> Type[Any] | Any:
        match self._annotation:
            case str():
                return self._evaluate_annotation()

            case _:
                return self._annotation

    def _evaluate_annotation(self) -> Type[Any] | Any:
        return eval(self._annotation, {}, EvaluationNamespace(self._some_namespace.value_or({})))


class EvaluationNamespace(Mapping[str, Any]):
    def __init__(self, namespace: dict[str, Any]):
        self.namespace = namespace

    def __getitem__(self, item: str) -> Any:
        if item in self.namespace:
            return self.namespace[item]

        match self.namespace.get("__builtins__", __builtins__):
            case dict() as builtins if item in builtins:
                return builtins[item]

            case builtins if hasattr(builtins, item):
                return getattr(builtins, item)

        return AnnotationForwardReference(item, Optional.Some(self.namespace))

    def __iter__(self):
        return iter(self.namespace)

    def __len__(self):
        return len(self.namespace)


class AnnotationForwardReference:
    def __init__(self, annotation: str, some_namespace: Optional[dict[str, Any]]):
        self._annotation = annotation
        self._some_namespace = some_namespace

    def __evaluate__(self):
        return Annotation(self._annotation, self._some_namespace)

    def __getattr__(self, name: str) -> Any:
        if name == "__typing_is_unpacked_typevartuple__":
            raise AttributeError(name)

        return AnnotationForwardReference(f"{self._annotation}.{name}", self._some_namespace)

    def __call__(self, *args, **kwargs):
        arg_string = ", ".join(map(repr, args))
        kwarg_string = ", ".join(f"{key}={value!r}" for key, value in kwargs.items())
        return AnnotationForwardReference(f"{self._annotation}({arg_string}, {kwarg_string})", self._some_namespace)

    def __getitem__(self, item: Any):
        return AnnotationForwardReference(f"{self._annotation}[{item!r}]", self._some_namespace)

    def __repr__(self):
        return f"<{type(self).__name__}: {self._annotation!r}>"



from typing import TypeVar, Generic

T = TypeVar("T")


class Container(Generic[T]):
    """Containers are used to provide a reference to a changeable value."""

    def __init__(self, default: T | None = None):
        self._value = default

    @property
    def value(self):
        return self._value

    def set(self, value: T) -> None:
        self._value = value
